# wigli wigli_bots.py

import openai

from copy import deepcopy
from dotenv import load_dotenv
from json import load
from os import getenv
from os.path import join
from slugify import slugify
from time import time
from typing import Any, Callable, Iterable, List, TYPE_CHECKING

from wigli._wigli_tools import (
    clamp,
    count_tokens,
    format_timestamp,
    plural,
    pluralize,
)

if TYPE_CHECKING:
    from wigli._wigli_data import WigliData

MAX_COMMANDS = 8
DEFAULT_TEMPERATURE = 1
MAX_TOKENS = 2**12
CH_PER_TOK = 4


def format_message(
    message: str | dict or "WigliMessage", role: str = "system"
) -> "WigliMessage":
    if isinstance(message, str):
        message = {"role": role, "content": message}
    if not isinstance(message, WigliMessage):
        message = WigliMessage(message)
    return message


def format_messages(
    messages: str
    | dict
    | Iterable[str | dict or "WigliMessage"]
    or "WigliMessage" = [],
    role="system",
) -> List["WigliMessage"]:
    """
    Takes messages in any compatible format, performs robust \
    type checking, and converts them to a list of WigliMessage objects
    """
    if (
        not isinstance(messages, Iterable)
        or isinstance(messages, str)
        or isinstance(messages, dict)
    ):
        messages = [messages]

    return [format_message(msg, role=role) for msg in messages]


def assert_type(
    subject: Any,
    name: str,
    expected_types: type | Iterable[type],
    error: Exception = TypeError,
    msg: str = None,
):
    if not isinstance(expected_types, Iterable):
        expected_types = [expected_types]
    for n, expected_type in enumerate(expected_types):
        # if not isinstance(expected_type, type):
        #     raise TypeError(
        #         f"expected_types must contain only elements of type type but the element at index {n} is of type {type(expected_type)}"
        #     )
        if isinstance(subject, expected_type):
            break
    else:
        if msg is not None:
            raise error(msg)
        else:
            raise error(
                f"""{name} must be of type {" or ".join(expected_types)}"""
            )


## Objects


class WigliMessage(object):
    def __init__(
        self,
        message: str | dict = None,
        role: str = None,
        timestamp: float = None,
    ):
        self.role = "user"
        self.content = ""
        self.timestamp = None

        if isinstance(message, dict):
            temp_role = message.get("role")
            if isinstance(temp_role, str):
                self.role = temp_role
            message = message.get("content")
        if isinstance(message, str):
            self.content = message
        if isinstance(role, str):
            self.role = role
        if isinstance(timestamp, float):
            self.timestamp = timestamp
        if self.timestamp is None:
            self.timestamp = time()

    def __str__(self) -> str:
        return f"""{self.role.title()}, {format_timestamp(self.timestamp)}: {self.content}"""

    def md(self, truncation: int = None):
        if isinstance(truncation, int):
            message = "> " + self.content[
                :truncation
            ].strip().replace("\n", "\n> ")
        else:
            message = "> " + self.content.strip().replace(
                "\n", "\n> "
            )
        return f"""\
#### [{self.role.title()}](##### "{format_timestamp(self.timestamp)}"):\n\n{message}"""

    def items(self):
        return [("role", self.role), ("content", self.content)]


class WigliInjection(object):
    def __init__(
        self,
        injection_messages: str
        | dict
        | WigliMessage
        | Iterable[WigliMessage | str | dict],
        reminder_messages: str
        | dict
        | WigliMessage
        | Iterable[WigliMessage | str | dict] = None,
        reminder_period: int = 5,
        set_timer: int = 0,
        role="system",
    ):
        self.injection_messages = format_messages(
            injection_messages,
            role=role,
        )

        assert_type(reminder_period, "reminder_timer", int)
        self.reminder_period = reminder_period
        assert_type(set_timer, "set_timer", int)

        self.reminder_timer = set_timer
        self.reminder_period = reminder_period
        self.reminder_messages = reminder_messages

        if self.reminder_messages is None:
            return

        self.reminder_messages = format_messages(
            self.reminder_messages
        )

    def do_reminder_tick(self):
        self.reminder_timer += 1
        if self.reminder_timer >= self.reminder_period:
            self.reminder_timer = 0
            return self.reminder_messages
        return []

    def __hash__(self):
        return hash(
            "\n".join(
                [str(msg) for msg in self.injection_messages]
            )
        )


class WigliCommand(WigliInjection):
    def __init__(
        self,
        cmd: dict,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        keyword = cmd.get("keyword")
        assert_type(keyword, "keyword", str)
        self.keyword = keyword
        run_function = cmd.get("run_function")
        assert_type(run_function, "run", Callable)
        self.run = run_function
        parse_function = cmd.get("parse_function")
        assert_type(parse_function, "parse", Callable)
        self.parse = parse_function


class WigliBot(object):
    TIMEOUT = 10

    LIMIT_MSG = "[ERROR: CONVERSATION LIMIT REACHED]"
    EMPTY_MSG = "[ERROR: NO PROMPT]"
    KEY_ERR_MSG = "[ERROR: NO OPENAI API KEY]"
    TIMEOUT_MSG = f"[ERROR: OPENAI FAILED TO RESPOND FOR {TIMEOUT} SECONDS]"

    def __init__(
        self,
        messages: str
        | dict
        | WigliMessage
        | Iterable[WigliMessage | dict]
        | WigliInjection
        | None = None,
        wiglidata: None or "WigliData" = None,
        prompt: str | None = None,
        title: str | None = None,
        filename: str | None = None,
        birthstamp: float = time(),
        touchstamp: float | None = None,
        stream: bool = True,
        reminders: set | None = set(),
    ):
        # Default class attribute values
        self.messages = []
        self.reminders = reminders
        self.stream = stream

        self.wiglidata = wiglidata
        if self.wiglidata is not None:
            self.log = self.wiglidata.log

        self.title = title
        self.filename = filename
        self.birthstamp = birthstamp
        self.touchstamp = (
            self.birthstamp
            if touchstamp is None
            else touchstamp
        )

        if openai.api_key == "":
            load_dotenv(dotenv_path=join(self.wiglidata.data_dir, ".env"))
            load_dotenv()
            openai.api_key = getenv("OPENAI_API_KEY")

        if messages is not None:
            self.Inject(messages)
        if prompt is not None:
            self.Inject(prompt)

    @classmethod
    def FromBot(cls, bot, *args, **kwargs):
        copied_attrs = {}
        for attr in vars(bot).keys():
            if attr != "log":
                value = getattr(bot, attr)
                copied_attrs[attr] = deepcopy(value)

        # Update copied_attrs with kwargs, overwriting any conflicting keys
        copied_attrs.update(kwargs)
        return cls(*args, **copied_attrs)

    # @classmethod
    # def FromJsonFile(cls, filename, *args, **kwargs):
    #     json_obj = {}
    #     with open(filename) as file:
    #         json_obj = load(file)
    #         bot = decode(json_obj)
    #         if isinstance(bot, WigliBot):
    #             return cls.FromBot(bot, *args, **kwargs)
    #     obj = cls(*args, **kwargs)
    #     obj.log(f"There was a problem loading {filename}", v=1)
    #     return obj

    @classmethod
    def FromFile(cls, filename, *args, **kwargs):
        json_obj = {}
        with open(filename) as file:
            json_obj = load(file)
            return cls.With(
                json_obj.get("messages", []), *args, **kwargs
            )
        obj = cls(*args, **kwargs)
        obj.log(f"There was a problem loading {filename}", v=1)
        return obj

    @classmethod
    def With(
        cls,
        prompt: str
        | dict
        | WigliMessage
        | Iterable[WigliMessage | dict]
        | WigliInjection = None,
        *args,
        **kwargs,
    ) -> "CommandBot":
        obj = cls(*args, **kwargs)
        if prompt is not None:
            obj.Inject(prompt)
        return obj

    def Inject(self, *args, **kwargs) -> "WigliBot":
        return self.And(*args, *kwargs)

    def And(
        self,
        messages: str
        | dict
        | WigliMessage
        | Iterable[WigliMessage | dict]
        | WigliInjection,
        role: str = "system",
    ) -> "WigliBot":
        if (
            isinstance(messages, Iterable)
            and len(messages) <= 0
        ):
            return self
        if not isinstance(messages, WigliInjection):
            messages = WigliInjection(messages, role=role)
        self.messages += messages.injection_messages
        if messages.reminder_messages is not None:
            self.reminders |= {messages}
        if self.wiglidata is not None:
            self.wiglidata.archive_chat(self)
        return self

    def make_filename(self):
        filename = slugify(
            self.title, ok="_", only_ascii=True, lower=False
        )
        self.filename = str(self.touchstamp) + "_" + filename

    def set_logger(self, logger: "WigliData"):
        self.wiglidata = logger
        self.log = logger.log

    def do_reminders_tick(self):
        for reminder in self.reminders:
            self.Inject(reminder.do_reminder_tick())

    def Chat(
        self,
        prompt: str
        | dict
        | WigliMessage
        | Iterable[WigliMessage | dict]
        | WigliInjection = None,
        stream: bool = None,
        model: str = "gpt-3.5-turbo",
        temperature: float = DEFAULT_TEMPERATURE,
        n: int = 1,
        stop: str | list = None,
        max_tokens: int = None,
        presence_penalty: float = 0,
        frequence_penalty: float = 0,
        logit_bias: dict = {},
        user: str = None,
        nochat: bool = False,
    ) -> str:
        if stream is None:
            stream = self.stream
        if prompt is not None:
            self.And(prompt, role="user")
        if nochat:
            return ""

        def dummy_message():
            dummy_message = "Let me think about that..."
            if stream:
                print(dummy_message)
            return dummy_message

        # return dummy_message() # Debug option for no API calls

        # Check if any messages exist
        if len(self.messages) <= 0:
            if stream:
                print(self.EMPTY_MSG, flush=True)
            return self.EMPTY_MSG

        # Check if messages will fit in model
        tokens = count_tokens(self.messages, model)
        self.log(
            f"Counted {tokens} {pluralize('token')} in self.messages"
        )
        if tokens > MAX_TOKENS:
            if stream:
                print(self.LIMIT_MSG, flush=True)
            return self.LIMIT_MSG

        # Convert _WigliMessages into dicts
        dict_messages = [
            {key: val for key, val in m.items()}
            for m in self.messages
        ]

        try:
            # OpenAI API request
            completion = openai.ChatCompletion.create(
                model=model,
                messages=dict_messages,
                temperature=temperature,
                # n=n,
                stream=stream,
                # stop=stop,
                # max_tokens=max_tokens,
                # presence_penalty=presence_penalty,
                # frequence_penalty=frequence_penalty,
                # logit_bias=logit_bias,
                # user=user,
            )
        except openai.error.AuthenticationError as e:
            self.log(
                f'OpenAI API Key AuthenticationError... Run "wigli --set-api-key <your_openai_api_key>" to fix this.\n\n{e}',
                v=0,
            )
            return self.KEY_ERR_MSG

        if not stream:
            finish_reason = completion.choices[0].finish_reason
            response = WigliMessage(
                completion.choices[0].message
            )
        else:
            finish_reason = "stop"
            response = WigliMessage(role="assistant")

            for event in completion:
                choice = event.choices[0]
                if "content" in choice.delta.keys():
                    response.content += choice.delta.content
                    print(
                        choice.delta.content, end="", flush=True
                    )
                finish_reason = choice.finish_reason

        if finish_reason == "length":
            if stream:
                print(self.LIMIT_MSG, flush=True)
            response.content += self.LIMIT_MSG
            return response.content

        if stream:
            print("")

        self.And(response)
        self.do_reminders_tick()
        return response.content

    def extract_transcript(
        self, limit=None, truncation=None, start=0
    ):
        archive_len = len(self.messages) - start
        if limit is not None and limit < archive_len - start:
            script_len = limit
        else:
            script_len = archive_len

        dates = []
        roles = []
        messages = []
        for n in range(
            archive_len - start - script_len,
            archive_len - start,
        ):
            message = self.messages[n]
            dates.append(format_timestamp(message.timestamp))
            roles.append(message.role)
            if truncation is not None:
                messages.append(
                    "> "
                    + message.content[:truncation]
                    .strip()
                    .replace("\n", "\n> ")
                )
            else:
                messages.append(
                    "> "
                    + message.content.strip().replace(
                        "\n", "\n> "
                    )
                )

        return dates, roles, messages

    def format_transcript(
        self,
        limit=None,
        truncation=None,
        linestart="",
    ):
        """
        Returns a transcript of the chat in plain text format.
        Each message is prefaced by the date, time, and sender"s name.

        Parameters
        ----------
        limit: int
            The amount of messages to limit the transcription to.
        truncation: int
            The amount of characters to limit each message to.
        linestart: str
            What to start each message with.

        Returns
        -------
        str
            The entire chat history in plain human-readable text.
        """

        archive_len = len(self.messages)
        self.log(f"archive_len is {archive_len}")
        if limit is not None and limit < archive_len:
            script_len = limit
        else:
            script_len = archive_len
        dates, roles, messages = self.extract_transcript(
            limit=limit
        )

        return "\n\n".join(
            [
                f"""\
{linestart}{dates[n]} [{roles[n]}]: \
{
    (
        (messages[n][:truncation] + '...')
        if truncation and len(messages[n]) > truncation
        else messages[n]
    ) 
}
"""
                for n in range(
                    archive_len - script_len, archive_len
                )
            ]
        )

    def format_transcript_markdown(self, limit=None):
        """
        Returns a transcript of the chat in markdown format.
        Each message is a blockquote with the date, time, and sender"s
        name in a level 5 header in the top of the blockquote.

        Parameters
        ----------
        limit: int
            The amount of messages to limit the transcription to.
        truncation: int
            The amount of characters to limit each message to.

        Returns
        -------
        str
            The entire chat history formatted in markdown.
        """

        dates, roles, messages = self.extract_transcript(
            limit=limit
        )

        return "\n\n".join(
            [
                f"""\
#### [{roles[n].title()}](##### "{dates[n]}"):\n\n{messages[n]}"""
                for n in range(len(messages))
            ]
        )

    def erase_messages(self, num_messages_to_erase):
        """
        Removes messages from the end of the chat and updates the birthstamp.

        Parameters
        ----------
        num_messages_to_erase: int
            How many messages to erase.
        """
        # Clamp the argument to valid values
        num_messages_to_erase = clamp(
            num_messages_to_erase, 0, len(self.messages)
        )
        # Only trim if the argument is nonzero
        if num_messages_to_erase > 0:
            self.log(
                (
                    "Trimming last",
                    num_messages_to_erase,
                    plural("message", num_messages_to_erase),
                    "from chat",
                )
            )
            self.messages = self.messages[
                :-num_messages_to_erase
            ]
            # Update birthstamp so original doesn't get deleted
            self.birthstamp = self.touchstamp

    def _last_message(self) -> str:
        if len(self.messages) > 0:
            return self.messages[-1].content
        return ""

    def log(self, s, end="\n", sep=" ", v=3):
        pass


class OneShotBot(WigliBot):
    def __init__(self, *args, **kwargs):
        stream = kwargs.get("stream", None)
        if stream is None:
            kwargs["stream"] = False
        self.run_args = kwargs.pop("run_args", {})
        assert_type(self.run_args, "run_args", dict)
        run_method = kwargs.pop("run_method", None)
        if run_method is not None:
            assert_type(run_method, "run_method", Callable)
            self.run = run_method
        super().__init__(*args, **kwargs)

    def __new__(cls, *args, **kwargs):
        instance = super().__new__(cls)
        instance.__init__(*args, **kwargs)
        return instance.run(instance.run_args)

    def run(self, run_args):
        return self.Chat()


class CommandBot(WigliBot):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.active_cmds = set()

    def Chat(
        self,
        prompt: str
        | dict
        | WigliMessage
        | Iterable[WigliMessage | dict]
        | WigliInjection = None,
        *args,
        **kwargs,
    ) -> str:
        reprompt = kwargs.pop("reprompt", True)
        nochat = kwargs.pop("nochat", False)
        if nochat:
            if prompt is not None:
                self.And(prompt, role="user")
            if len(self.messages) < 1:
                return ""
            response = self.messages[-1].content
        else:
            response = super().Chat(prompt, *args, **kwargs)

        args = []
        for num_commands in range(MAX_COMMANDS):
            self.log("Parsing message for commands")
            for cmd in self.active_cmds:
                if cmd.keyword in response:
                    arg = cmd.parse(response)
                    if arg is not None:
                        arg["stream"] = self.stream
                        result = cmd.run(arg)
                        break
            else:
                break

            if result[-1].role == "quit":
                break

            if not reprompt:
                break

            self.log("Reprompting with command output")
            kwargs["prompt"] = result

            response = super().Chat(*args, **kwargs)
        else:
            print(
                f"""\
The maximum allowed number of bot commands in a row is {MAX_COMMANDS}, \
but Wigli would like to keep going. To allow {MAX_COMMANDS} more \
bot {plural('command', MAX_COMMANDS)}, use the -b flag."""
            )

        return self.messages[-1].content

    def And(
        self,
        messages: str
        | dict
        | WigliMessage
        | Iterable[WigliMessage | dict]
        | WigliInjection
        or WigliCommand,
        role: str = "system",
    ) -> "CommandBot":
        if (
            isinstance(messages, dict)
            and messages.get("parse_function", None) is not None
            and messages.get("injection_messages", None)
            is not None
        ):
            messages = WigliCommand(
                messages, messages.get("injection_messages", [])
            )
        if isinstance(messages, WigliCommand):
            if messages not in self.active_cmds:
                super().And(messages)
            self.active_cmds |= {messages}
        elif messages is not None:
            super().And(messages)
        return self



# wigli _wigli_argparser.py

from argparse import ArgumentParser, RawTextHelpFormatter
from time import time

from wigli._wigli_tools import contains_any, format_timestamp
from wigli._wigli_version import VERSION

MAX_VERBOSITY = 3


def contains_any_flagargs(list, flags):
    for f in list:
        if f[0] == "-" and f[1] != "-":
            for flag in flags:
                if flag[1:] in f:
                    return True
    return False


def fetch_args(argv):
    no_prompt_args = [
        "--noprompt",
        # "--transcript",
        "--transcript-full",
        "--list",
        "--list-installed-plugins",
        "--list-plugins",
        "--version",
        "--botcommands",
        "--set-api-key",
        # "--audio",
    ]
    no_prompt_flags = [
        "-m",
        # "-t",
        "-T",
        "-l",
        "-V",
        "-b",
        "-L",
        "-K",
        # "-a",
    ]

    parser = ArgumentParser(
        prog="wigli",
        description=f"""\
      )  (              Wigli v{VERSION} - {format_timestamp(time())} its like ChatGPT but better!
     (   ) )
   (     (    )         Wigli is an extensible open source framework and CLI
 (____(___)_____)___    designed to support the development of SDX chatbots,
 |              |---\\\\  datamining botswarms, and other bleeding-edge AI tools.
 |              |    ||
 |              |___//  Wigli can already use DuckDuckGo, URL summarization and
  \            /----'   Python programming to better assist the user, with Voice
   \          /         and WebUI coming soon. Your chat history is saved locally,
    \________/          and you can use your own OpenAI API key. Happy chatting!""",
        epilog="""\
You can contribute new injections and extensions for Wigli!
Visit https://github.com/VulcanicAI/Wigli to learn more.
""",
        # conflict_handler="resolve",
        formatter_class=RawTextHelpFormatter,
    )

    parser_parse_cmds = parser.add_mutually_exclusive_group()
    parser_resume = parser.add_mutually_exclusive_group()

    if (
        "--fileprompt" in argv
        or contains_any_flagargs(argv, ["-f"])
        or not (
            contains_any(argv, no_prompt_args)
            or contains_any_flagargs(argv, no_prompt_flags)
        )
    ):
        parser.add_argument(  # Only require prompt if necessary
            "prompt",
            type=str,
            default="",
            help="the prompt to submit",
        )
    else:
        parser.set_defaults(prompt="")

    parser.add_argument(
        "-v",
        "--verbosity",
        action="count",
        default=0,
        help=f"verbosity with which to print logs, cumulative up to {MAX_VERBOSITY}",
    )
    parser.add_argument(
        "-V",
        "--version",
        action="store_true",
        help=f"print version number",
    )
    parser.add_argument(
        "-K",
        "--set-api-key",
        metavar="API_KEY",
        type=str,
        help="Create a .env file with your OpenAI API Key to enable ChatGPT access",
    )
    parser.add_argument(
        "-l",
        "--list",
        action="store_true",
        help="list past conversations",
    )
    # parser.add_argument(
    #     "-n",
    #     "--num",
    #     type=int,
    #     default=None,
    #     help="the number of previous conversations to list",
    # )
    # parser.add_argument(
    #     "-p",
    #     "--preview",
    #     type=int,
    #     default=0,
    #     help="how many messages to preview in each list entry",
    # )
    parser.add_argument(
        "-c",
        "--clean",
        action="store_true",
        help="force a fresh instance",
    )
    parser.add_argument(
        "-s",
        "--search",
        action="store_true",
        help="enable assistant to use search commands",
    )
    parser_resume.add_argument(
        "-r",
        "--resume-last",
        action="store_true",
        help="resume previous chat session",
    )
    parser_resume.add_argument(
        "--time",
        type=str,
        default=None,
        help="supply the timestamp of a previous conversation to resume",
    )
    parser_resume.add_argument(
        "-i",
        "--index",
        type=int,
        default=None,
        help="supply the chronological index of a previous conversation to resume",
    )
    parser_resume.add_argument(
        "--title",
        type=str,
        default=None,
        help="supply the title of a previous conversation to resume",
    )
    # parser.add_argument(
    #     "-t",
    #     "--transcript",
    #     action="store_true",
    #     help="no chat just print conversation transcript with one-line previews",
    # )
    parser.add_argument(
        "-T",
        "--transcript-full",
        action="store_true",
        help="no chat just print conversation transcript in full",
    )
    parser.add_argument(
        "--system",
        type=str,
        help="optional system prompt before user prompt for behavior modification",
    )
    parser.add_argument(
        "-m",
        "--noprompt",
        action="store_true",
        help="just complete the next chat based on the conversation history",
    )
    parser_parse_cmds.add_argument(
        "-b",
        "--botcommands",
        action="store_true",
        help="parse the last message for bot commands first",
    )
    parser_parse_cmds.add_argument(
        "-u",
        "--usercommands",
        action="store_true",
        help="No chat, just parse the last message for commands",
    )
    parser.add_argument(
        "-e",
        "--erase",
        type=int,
        help="erase a given number of messages from the conversation history",
    )
    parser.add_argument(
        "-x",
        "--python",
        action="store_true",
        help="enable python interpreter",
    )
    parser.add_argument(
        "-P",
        "--professor",
        action="store_true",
        help="summon the professor",
    )
    parser.add_argument(
        "-f",
        "--fileprompt",
        action="store_true",
        help="prompt will be read from file",
    )
    # parser.add_argument(
    #     "-D",
    #     "--datadir",
    #     type=str,
    #     help="specify a directory to save and load Wigli chat history",
    # )
    # parser.add_argument(
    #     "-a",
    #     "--audio",
    #     action="store_true",
    #     help="prompt will be transcribed from audio",
    # )
    parser.add_argument(
        "--doctest",
        action="store_true",
        help=f"interpret prompt as a Python function and invoke the headless DocTest bot",
    )
    # parser.add_argument(
    #     "-L",
    #     "--list-installed-plugins",
    #     action="store_true",
    #     help=f"list installed plugins",
    # )
    # parser.add_argument(
    #     "--list-plugins",
    #     action="store_true",
    #     help=f"list available plugins on PyPI",
    # )

    return parser.parse_args(argv)



# wigli _wigli_cli.py

from jsonpickle import decode
from math import floor
from os.path import join
from sys import version_info, argv
from time import time
from typing import List

from wigli import WigliBot, CommandBot, WigliMessage
from wigli._wigli_argparser import fetch_args
from wigli._wigli_version import VERSION
from wigli._wigli_data import WigliData
from wigli._wigli_tools import clamp

# if version_info < (3, 10):
#     from importlib_metadata import entry_points
# else:
#     from importlib.metadata import entry_points

from wigli._wigli_plugins import (
    Doctest,
    HistoryProfessor,
    PythonBot,
    SearchBot,
)


def wigli_cli(
    argv_: List[str] | None = argv[1:],
    data_dir: str | None = None,
    verbosity: int = 0,
):
    """
    This is the main entry point to the program from the
    command-line. It uses sys.argv to create and run an instance of
    WigliInvocation in accordance with the user's command-line arguments.
    """

    invocation = WigliInvocation(
        argv_, data_dir=data_dir, verbosity=verbosity
    )
    if invocation.prompt is None:
        return

    invocation.log("Starting chat")
    invocation.Chat()


class WigliInvocation(object):
    def __init__(
        self,
        argv: List[str],
        data_dir: str | None = None,
        verbosity: int = 0,
    ):
        self.archive = None
        self._timestamp = time()

        self.data = WigliData(
            self._get_timestamp,
            data_dir=data_dir,
            verbosity=verbosity,
        )

        self.log = self.data.log
        self.log("Hello, logger!")
        self.bot = None
        self.argv = argv
        self.args = fetch_args(argv)
        self.prompt = self._handle_args()
        self.log(f"fetched args for argv: {argv}")

    def Chat(self) -> str:
        return self.bot.Chat(self.prompt)

    def _get_timestamp(self):
        return self._timestamp

    def _bot_from_args(self) -> str:
        injection = []

        if self.args.system is not None:
            injection = [
                WigliMessage(self.args.system, "system")
            ] + injection

        if self.bot is None:
            self.bot = WigliBot(wiglidata=self.data)

        if self.args.professor:
            self.log("Summoning Professor Wigli", v=0)
            self.bot = HistoryProfessor.FromBot(self.bot)

        if self.args.python:
            self.log("Injecting Python capabilities", v=0)
            self.bot = PythonBot.FromBot(self.bot)

        if self.args.search:
            self.log("Injecting web search capabilities", v=0)
            self.bot = SearchBot.FromBot(self.bot)

        self.bot.Inject(injection)

        return self.args.prompt

    def _handle_args(self):
        """
        Handles args and modifies the state of this instance accordingly.
        """

        self.data.verbosity = self.args.verbosity

        if self.args.version:
            print(f"You are using Wigli v{VERSION}")
            return
    
        if self.args.set_api_key is not None:
            with open(join(self.data.data_dir, ".env"), "w") as f:
                f.write(
                    f"""\
# Once you add your API key below, make sure to not share it with anyone! The API key should remain private.
OPENAI_API_KEY={self.args.set_api_key}
"""
                )
            self.log(f"Saved API key at {join(self.data.data_dir, '.env')}", v=0)
            return
        
        # if self.args.list_installed_plugins:
        #     discovered_plugins = entry_points(group='wigli._wigli_plugins')
        #     for discovered_plugin in discovered_plugins:
        #         print(f"    {discovered_plugin}")
        #     return

        # List previous chats
        if self.args.list:  # or self.args.transcriptbrief:
            self.log(
                "self.args.list was True, printing history then exiting"
            )
            self.data.print_chat_history_oneline()
            return

        # Load previous chat
        self.bot = None
        if not self.args.clean:
            self.bot = self._load_chat()
            if isinstance(self.bot, WigliBot):
                self.log(
                    f"""
Loaded previous chat: {getattr(self.bot, "title", "No title")}""",
                )
                self.bot.wiglidata = self.data
                self.bot.log = self.data.log
                if self.args.erase is not None:
                    self.bot.erase_messages(self.args.erase)
            else:
                self.bot = None

        # Now everything is initialized

        # No chat, just print transcripts
        if self.args.transcript_full:
            self.log("Printing transcript then exiting")
            if self.bot is not None:
                print(self.bot.format_transcript())
            else:
                print("No chat for which to print transcript")
            return

        # Parse the last message for commands and prompt
        if self.args.botcommands and isinstance(
            self.bot, CommandBot
        ):
            self.log(
                "Parsing last message for commands, then reprompting with output"
            )
            self.bot.Chat(nochat=True)
            return

        # # # if self.args.audio:
        # # #     if self.args.fileprompt:
        # # #         print("Loading audio prompt from file")
        # # #     else:
        # # #         self.args.prompt = record_prompt()
        # # #     prompt = ""
        # # #     with open(self.args.prompt, "rb") as f:
        # # #         prompt = Audio.transcribe("whisper-1", f).text
        # # #     self.args.prompt = prompt
        # # #     if self.args.prompt != "":
        # # #         print(f"Heard:\n{self.args.prompt}")
        # # # elif self.args.fileprompt:  # Load prompt from file

        if self.args.fileprompt:  # Load prompt from file
            filepath = self.args.prompt
            self.args.prompt = ""
            with open(filepath) as f:
                self.log(
                    ("Reading text prompt from file:", filepath)
                )
                self.args.prompt = f.read()

        # Treat the prompt as code, invoke the headless doctest bot, don't prompt
        if self.args.doctest:
            self.log(
                "Invoking the headless DocTest bot on the following code:\n\n{self.args.prompt}"
            )
            self.log(Doctest(self.args.prompt), v=0)
            return

        prompt = self._bot_from_args()

        if self.bot is None:
            self.log(
                "No previous chat to load, starting a new one"
            )
            self.bot = WigliBot(wiglidata=self.data)

        # Parse the last message for commands and don't prompt
        if self.args.usercommands and isinstance(
            self.bot, CommandBot
        ):
            self.log(
                "Parsing last message for commands, then exiting"
            )
            self.bot.Chat(
                self.args.prompt, nochat=True, reprompt=False
            )
            return

        return prompt

    def _load_chat(self):
        """
        Searches the archive for a valid chat
        to load based on the arguments provided

        Returns
        -------
        dict
            The dict object that was loaded, or None
        """
        logs = self.data.list_chats()

        if len(logs) == 0:
            return

        if self.args.resume_last:
            # Load the previous conversation
            with open(
                join(
                    self.data.convos_dir,
                    logs[len(logs) - 1],
                )
            ) as f:
                return decode(f.read())

        if self.args.index is not None:
            # Load a previous conversation by reverse chronological index
            self.args.index = clamp(
                self.args.index, 1, len(logs)
            )
            with open(
                join(
                    self.data.convos_dir,
                    logs[len(logs) - self.args.index],
                )
            ) as f:
                return decode(f.read())

        if self.args.time is not None:
            # Load a previous conversation by timestamp
            files = [
                log for log in logs if self.args.time in log
            ]
            if len(files) > 0:
                with open(
                    join(
                        self.data.convos_dir,
                        files[len(files) - 1],
                    )
                ) as f:
                    return decode(f.read())

        if self.args.title is not None:
            # Load a previous conversation by title
            files = [
                log
                for log in logs
                if self.args.title.replace(" ", "").lower()
                in log.replace("_", "").lower()
            ]
            if len(files) > 0:
                with open(
                    join(
                        self.data.convos_dir,
                        files[len(files) - 1],
                    )
                ) as f:
                    return decode(f.read())

        # Check for chat in the last 5 minutes
        max_timestamp = floor(float(logs[0].split("_")[1]))
        max_index = 0
        for i in range(1, len(logs)):
            timestamp = floor(float(logs[i].split("_")[1]))
            if timestamp > max_timestamp:
                max_timestamp = timestamp
                max_index = i

        if self._timestamp - max_timestamp < 300:
            # Time since chat is fewer than 5 minutes
            with open(
                join(self.data.convos_dir, logs[max_index])
            ) as f:
                return decode(f.read())



# wigli __init__.py

from wigli._wigli_bots import (
    WigliMessage,
    WigliInjection,
    WigliCommand,
    WigliBot,
    OneShotBot,
    CommandBot,
)

from wigli._wigli_cli import WigliInvocation, wigli_cli

NOTES = f"""\
This command-line toolset implements the \
OpenAI API into several tools and bots.

For an up-to-date-list, run "ch -h". This \
header documents system conventions and to-dos.

TODO: Better titling
TODO: Assistant pip access
TODO: Automated tests
TODO: Search bot
TODO: Auto-git botswarm
    Commit decider, commit titler, FracTLDR_Code
TODO: transcriptbrief title search and timestamp search integration
TODO: AI-powered manual for this CLI
TODO: Ensure entire dependency tree is MIT license
TODO: Refactor history to inherit from OpenAI objects
TODO: Break out args handler
TODO: Break out command parser
TODO: Diagram for comparison to other assistants
TODO: generate_and_append_reply do_parse_cmds parameter seems hacky
TODO: Stacktrace summarizer
TODO: Automatic forgetting
TODO: Key input
TODO: ChatCompletion timeout watchdog
TODO: Chai house multi-chat protocol
TODO: handll house
\
"""



# wigli _wigli_data.py

from appdirs import user_data_dir
from jsonpickle import encode, decode
from os.path import abspath, join
from time import time
from typing import Callable

from wigli import WigliBot, OneShotBot

from wigli._wigli_tools import (
    list_dir,
    make_dir,
    remove_file,
)


class WigliData(object):
    """
    This class handles chat data for a WigliInvocation and its WigliBots.

    Attributes
    ----------
    get_time: Callable
        WigliData uses this callback function to name its log files.
    data_dir: str
        The directory for loading and saving chat history.
    convos_dir: str
        Beneath data_dir, stores full JSON chat history and data files.
    scripts_dir: str
        Beneath data_dir, stores pretty human-readable transcripts of chats.
    logs_dir: str
        Beneath data_dir, stores event log files.
    verbosity: int, optional
        Level of verbosity at which to log events.

    Methods
    -------
    __init__()
        Sets up the user data directory.
    list_chats()
        Returns all files in the
    print_chat_history_oneline()
        Prints a numbered list of conversations in the archive.
    log()
        Event logger with a range of verbosities.
        Always log to file and sometimes log to console too.
    """

    def __init__(
        self,
        get_time: Callable,
        data_dir: str | None = None,
        verbosity: int = 0,
    ):
        """
        Sets up the user data directory.

        Parameters
        ----------
        get_time: Callable
            WigliData uses this callback function to name its log files.
        data_dir: str, optional
            The directory for loading and saving chat history (defaults to appdirs.user_data_dir).
        verbosity: int, optional
            Level of verbosity at which to log events.
        """

        self.get_time = get_time
        self.verbosity = verbosity
        if data_dir is not None:
            self.data_dir = data_dir
        else:
            self.data_dir = user_data_dir(
                appname="Wigli", appauthor="VulcanicAI"
            )

        self.convos_dir = join(self.data_dir, "Wigli Files")
        self.scripts_dir = join(self.data_dir, "Pretty Chats")
        self.logs_dir = join(self.data_dir, "Debug Logs")

        make_dir(self.data_dir)
        make_dir(self.convos_dir)
        make_dir(self.scripts_dir)
        make_dir(self.logs_dir)

    def list_chats(self, suffix=".json"):
        """
        Returns a list of all chats in the archive.

        Parameters
        ----------
        suffix: str
            The pattern to filter by when listing the files.

        Returns
        -------
        list
            A list of all the files in the Wigli
            Data directory, in chronological order
            (because the filenames start with their touchstamps)
        """
        return sorted(
            [
                log
                for log in list_dir(self.convos_dir)
                if suffix in log
            ]
        )

    def print_chat_history_oneline(self, begin=0, end=None):
        """
        Prints a numbered list of conversations in the archive.

        Parameters
        ----------
        begin: int
            How many recent chats to skip.
        end: int
            How far back to go.
        """

        logs = self.list_chats()

        if end is None:
            end = len(logs)

        for n in range(end, begin, -1):
            with open(
                join(
                    self.convos_dir,
                    logs[len(logs) - n],
                )
            ) as json_file:
                chat = decode(json_file.read())
                if len(getattr(chat, "messages", [])) > 0:
                    print(
                        f"{n}: {getattr(chat, 'title', 'No Title')}\n",
                        sep="",
                        end="",
                    )

    def archive_chat(self, bot: WigliBot):
        if len(bot.messages) <= 0:
            return

        old_filename = bot.filename
        overwrite = False
        if old_filename is not None:
            with open(
                join(
                    self.convos_dir,
                    "chat_" + old_filename + ".json",
                )
            ) as f:
                old_archive = decode(f.read())
                assert isinstance(old_archive, WigliBot)
                if old_archive.birthstamp == bot.birthstamp:
                    overwrite = True

        bot.touchstamp = time()
        self.write_files(bot)
        if overwrite:
            remove_file(
                join(
                    self.convos_dir,
                    "chat_" + old_filename + ".json",
                )
            )
            remove_file(
                join(
                    self.scripts_dir,
                    "transcript_" + old_filename + ".md",
                )
            )

    def open_file(
        self, filepath, open_type="r", encoding="utf-16"
    ):
        try:
            return open(filepath, open_type, encoding=encoding)
        except BaseException:
            self.log(("Couldn't open file:", filepath))
            pass

    def title_transcript(self, transcript):
        injection_transcript_titler_bot = [
            {
                "role": "system",
                "content": """\
You are an expert at summarizing the purpose of a conversation in just a few words. The user will ask you to summarize a conversation and give you the transcript. You will only use about two to four words but the core themes of the conversation will be clearly communicated. If you understand, please respond with a summary of this conversation we are having in the style of these summaries. The next message will be the user's transcript.""",
            },
            {
                "role": "assistant",
                "content": "Titling Transcripts",
            },
            {
                "role": "user",
                "content": f"""\
Here is the transcript I would like you to title. Please just respond with the most concise title and nothing else.\n\n{transcript}""",
            },
        ]

        return OneShotBot(
            messages=injection_transcript_titler_bot,
        )

    def write_files(self, bot: WigliBot):
        script = bot.format_transcript_markdown()
        if bot.title is None:
            self.log("Auto-titling transcript")
            bot.title = self.title_transcript(script)
            self.log(("Auto-titled:", bot.title))
        bot.make_filename()
        json_filename = "chat_" + bot.filename + ".json"
        json_filepath = join(self.convos_dir, json_filename)
        abspath_json_filepath = abspath(json_filepath)
        with open(
            abspath_json_filepath,
            "w",
        ) as f:
            self.log(
                (
                    "Saving chat JSON file at",
                    abspath_json_filepath,
                )
            )
            f.write(encode(bot, f, indent=4))
        txt_filename = "transcript_" + bot.filename + ".md"
        txt_filepath = join(self.scripts_dir, txt_filename)
        with self.open_file(txt_filepath, open_type="w") as f:
            self.log(
                ("Saving transcript text file at", txt_filepath)
            )
            try:
                f.writelines([script])
            except BaseException:
                with self.open_file(
                    txt_filepath,
                    open_type="w",
                    encoding="utf-32",
                ) as f:
                    try:
                        f.writelines([script])
                    except BaseException:
                        self.log(
                            (
                                "Failed to write script: \n",
                                script,
                            ),
                            sep="",
                        )
                        pass
                pass

    def log(self, s, end="\n", sep=" ", v=3):
        """
        Event logger with a range of verbosities.
        Always log to file and sometimes log to console too.

        Parameters
        ----------
        s: str
            The message to log.
        end: str, optional
            What to end the string with.
        sep: str, optional
            What to separate tokens in the tuple with, if applicable.
        v: int, optional
            The level of verbosity of this message.
        """

        if type(s) == tuple:
            s = "".join([str(f) + sep for f in s])
        elif type(s) != str:
            s = str(s)

        s += end

        prefix = ""
        if v == 1:
            prefix = "LOG: "
        if v == 2:
            prefix = "Log: "
        if v == 3:
            prefix = "log: "

        # log = "LOG: "[: -v - 2] + "log: "[-v - 2 :]
        s = prefix + s

        if v <= self.verbosity:
            print(s, end="")
        with open(
            join(
                self.logs_dir,
                "log_" + str(self.get_time()) + ".log",
            ),
            "a",
        ) as f:
            f.write(s)
